from enum import Enum


class ScorerName(str, Enum):
    agentic_session_success = "agentic_session_success"
    agentic_workflow_success = "agentic_workflow_success"
    bleu = "bleu"
    chunk_attribution_utilization = "chunk_attribution_utilization"
    completeness = "completeness"
    context_adherence = "context_adherence"
    context_relevance = "context_relevance"
    correctness = "correctness"
    ground_truth_adherence = "ground_truth_adherence"
    input_pii = "input_pii"
    input_sexist = "input_sexist"
    input_tone = "input_tone"
    input_toxicity = "input_toxicity"
    instruction_adherence = "instruction_adherence"
    output_pii = "output_pii"
    output_sexist = "output_sexist"
    output_tone = "output_tone"
    output_toxicity = "output_toxicity"
    prompt_injection = "prompt_injection"
    prompt_perplexity = "prompt_perplexity"
    rouge = "rouge"
    tool_error_rate = "tool_error_rate"
    tool_selection_quality = "tool_selection_quality"
    uncertainty = "uncertainty"
